/* particlesystem.cpp
Particlesystem code.
*/

//// Includes - System

#include <windows.h>
#include <mmsystem.h>
#include <GL/gl.h>
#include <GL/glaux.h>

//// Includes - Project
#include "glext.h"      // OpenGL Extentionfile, needed for OpenGL ARB Point Sprites
#include "particlesystem.h"


//// Function Pointers for extentions
// GL_ARB_point_parameters
PFNGLPOINTPARAMETERFARBPROC  glPointParameterfARB  = NULL;
PFNGLPOINTPARAMETERFVARBPROC glPointParameterfvARB = NULL;

ParticleSystem::ParticleSystem()
{
	checkExtentions();
	pTexture =-1;
	pSysAge = 0.0f;
	pSysLastItAge = 0.0f;
	activeParticles = 0;
	partActiveList = NULL;
	partFreeList = NULL;
	boxDistribution = false;
	fadeFact = 0.0f;

	pSysRndOn = false;

	maxSize = 0.0f;
    glGetFloatv( GL_POINT_SIZE_MAX_ARB, &maxSize );
}

ParticleSystem::~ParticleSystem()
{
	// Delete the texture
	glDeleteTextures( 1, &pTexture );

	// Loop through our particle-lists and delete the structs.
	while( partActiveList )
    {
        Particle* part = partActiveList;
        partActiveList = part->pNext;
        delete part;
    }
    partActiveList= NULL;

	while( partFreeList )
    {
        Particle* part = partFreeList;
        partFreeList = part->pNext;
        delete part;
    }
    partActiveList= NULL;
}

void ParticleSystem::setGravity(vector3f in_grav )
{
	pGrav = in_grav;
}

void ParticleSystem::setWind( vector3f in_wind )
{
	pWind = in_wind;
}

void ParticleSystem::setWindState(bool power)
{
	pWindOn = power;
}
void ParticleSystem::setInitialVelocity ( vector3f in_initVel )
{
	pInitVel = in_initVel;
}

void ParticleSystem::setParticleSize(float size)
{
	pSize = size;
}

void ParticleSystem::setRandomFactor(float perc)
{
	pSysRandomFactor = perc;
}

void ParticleSystem::setRandomOn(bool power)
{
	pSysRndOn = power;
}
void ParticleSystem::setBoxCorner(vector3f in_initPos)
{
	pBoxCorner = in_initPos;
	boxDistribution = true;
}

void ParticleSystem::setInitialPosition ( vector3f in_initPos )
{
	pInitPos = in_initPos;
}

void ParticleSystem::setReleaseInterval ( float in_lim )
{
	pRelInterval = in_lim;
}

void ParticleSystem::setReleaseNumber(int in_num)
{
	pRelNum = in_num;
}

void ParticleSystem::setParticleCount(int in_partnum)
{
	numParticles = in_partnum;
}
void ParticleSystem::setParticleLife(float in_partlife)
{
	pTTL = in_partlife;
}

void ParticleSystem::setFade(float perc)
{
	fadeFact = perc;
}

void ParticleSystem::setTexture ( char *in_texturename )
{
	AUX_RGBImageRec *pTextureImage = auxDIBImageLoad( in_texturename );

    if( pTextureImage != NULL )
	{
        glGenTextures( 1, &pTexture );

		glBindTexture( GL_TEXTURE_2D, pTexture );

		glTexParameteri( GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR );

		glTexImage2D( GL_TEXTURE_2D, 0, 3, pTextureImage->sizeX, pTextureImage->sizeY, 0,
				     GL_RGB, GL_UNSIGNED_BYTE, pTextureImage->data );
	}

	if( pTextureImage )
	{
		if( pTextureImage->data )
			free( pTextureImage->data );

		free( pTextureImage );
	}
}

void ParticleSystem::init()
{
	int initPartCount = (int)((pTTL / pRelInterval) * (float)pRelNum);

	for (; initPartCount > 0; initPartCount--)
	{
		Particle *partnew = new Particle;
		partnew->pNext = partFreeList;

		partFreeList = partnew;
	}
}

void ParticleSystem::update(double in_timePassed)
{
	// Take a fraction of the deltaTime in float so we don't get to big jumps in our system.
	float delta = (float)(in_timePassed) * 0.001f;

	// We use part for working and 
	Particle  *part;
    Particle **pointPart;

	// Update the particle systems age
    pSysAge += delta;

    pointPart = &partActiveList;

    while( *pointPart )
    {
        part = *pointPart; // Set a pointer to the head

        // How old is the current particle
        part->age += delta;

		// Have this particle gone over it's time to live
        if( part->age >= pTTL )
        {
			// Move the now free particle to the head of the freeList and
			// set up the next particle for the next iteration of the loop
            *pointPart = moveToFreeList(part);
        }
        else
        {
            // Update particle position and velocity

            // Update the velocity-vector
            part->vel += pGrav * delta;

            if(pWindOn)
                part->vel += pWind * delta;

            // Update the position
            part->pos += part->vel * delta;

			// Set up the next particle for the next iteration of the loop
            pointPart = &part->pNext;
        }
    }

	// Release new particles if allowed by the release-factors.

    if( pSysAge - pSysLastItAge > pRelInterval )
    {
        // Set last update time
        pSysLastItAge = pSysAge;
    
        // Release the burst of particles
        for( int i = 0; i < pRelNum; i++ )
        {
            // Are the any free particles
            if( partFreeList )
            {
                // Take the first available
                part = partFreeList;

                // Update the freelist
                partFreeList = part->pNext;
            }
            else
            {
				// No particles in the freelist, can we create a new?
                if( activeParticles < numParticles )
                    part = new Particle;
            }

			// Check to see if we can create a new particle in the activeList without passing the max allowed particles
            if( activeParticles < numParticles )
            {

				// Move the current particle to the front of our activeParticle list and update the list
                part->pNext = partActiveList;
                partActiveList = part;
				
				initParticle(part);
                
                activeParticles++;
            }
        }
    }

}


void ParticleSystem::render()
{
	// We'll be looping through our active-particle list.
	Particle  *part = partActiveList;

    // quadratic have the coefficients of the attenuation function that decides how the particles will scale as they move
	// into the distance
    float quadratic[] =  { 1.0f, 0.0f, 0.01f };
    glPointParameterfvARB( GL_POINT_DISTANCE_ATTENUATION_ARB, quadratic );

    // The alpha of a point is calculated to allow the fading of points 
    // instead of shrinking them past a defined threshold size. The threshold 
    // is defined by GL_POINT_FADE_THRESHOLD_SIZE_ARB and is not clamped to 
    // the minimum and maximum point sizes.
    glPointParameterfARB( GL_POINT_FADE_THRESHOLD_SIZE_ARB, 60.0f );

    glPointParameterfARB( GL_POINT_SIZE_MIN_ARB, 1.0f );
    glPointParameterfARB( GL_POINT_SIZE_MAX_ARB, maxSize );

	// Clamp the userdefined size to max-size

	if (pSize > maxSize)
		pSize = maxSize;

    // Specify point sprite texture coordinate replacement mode for each texture unit
    glTexEnvf( GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_TRUE );


	// Set the texture we're going to use
	glBindTexture( GL_TEXTURE_2D, pTexture );

    //
	// Render point sprites...
	//

    glEnable( GL_POINT_SPRITE_ARB );

    glPointSize( pSize );
	
	float alpha = 0.8f;

	glBegin( GL_POINTS );
    {
		// Render each particle...
		while( part )
		{
			if (fadeFact != 0.0f)
			{
				alpha = 1 - ((part->age) / ( pTTL * fadeFact));
			}

			glColor4f( 0.6f, 0.6f,0.6f, alpha);

			glVertex3f( part->pos.x,
					    part->pos.y,
					    part->pos.z );

			part = part->pNext;
		}
    }
	glEnd();

	glDisable( GL_POINT_SPRITE_ARB );
}

inline Particle* ParticleSystem::moveToFreeList(Particle *part)
{
	Particle *temp = part->pNext;
	part->pNext = partActiveList;
	--activeParticles;
	return temp;
}

// Give us a random float from min to max.
float ParticleSystem::rndFloatBetween( float min, float max )
{
    float rndNum = (float)rand () / RAND_MAX;
    return min + (max - min) * rndNum;
}

// When we create a new particle, this is what is needs to be set.
inline void ParticleSystem::initParticle(Particle *in_part)
{
	if (boxDistribution)
	{
		in_part->pos = vector3f(rndFloatBetween(pInitPos.x, pBoxCorner.x),
								rndFloatBetween(pInitPos.y, pBoxCorner.y),
								rndFloatBetween(pInitPos.z, pBoxCorner.z));
	} else {
		in_part->pos = pInitPos;
	}
	if (!pSysRndOn)
		in_part->vel = pInitVel;
	else {
		in_part->vel.x = pInitVel.x + (pInitVel.x * rndFloatBetween(-pSysRandomFactor, pSysRandomFactor));
		in_part->vel.y = pInitVel.y + (pInitVel.y * rndFloatBetween(-pSysRandomFactor, pSysRandomFactor));
		in_part->vel.z = pInitVel.z + (pInitVel.z * rndFloatBetween(-pSysRandomFactor, pSysRandomFactor));
	}
	in_part->age = 0;
}

// Check to see if the system supports the GL_ARB Point Sprite Extention
bool ParticleSystem::checkExtentions()
{
    char *ext = (char*)glGetString( GL_EXTENSIONS );

    if( strstr( ext, "GL_ARB_point_parameters" ) != NULL )
    {
        glPointParameterfARB  = (PFNGLPOINTPARAMETERFARBPROC)wglGetProcAddress("glPointParameterfARB");
        glPointParameterfvARB = (PFNGLPOINTPARAMETERFVARBPROC)wglGetProcAddress("glPointParameterfvARB");
        if( !glPointParameterfARB || !glPointParameterfvARB )
        {
            MessageBox(NULL,"One or more GL_ARB_point_parameters functions were not found",
                "ERROR",MB_OK|MB_ICONEXCLAMATION);
        }
		return false;
    }
    else
    {
		MessageBox(NULL,"GL_ARB_point_parameters extension was not found",
            "ERROR",MB_OK|MB_ICONEXCLAMATION);
		return false;
    }
	return true;
}