/* main.cpp
Implementering av partikkelsystemet for Grafisk Databehandling
Aleksander Toppe
*/

//// Defines
#define STRICT
#define WIN32_LEAN_AND_MEAN

//// Includes - System

#include <windows.h>
#include <mmsystem.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glaux.h>

//// Includes - Project
#include "resource.h"
#include "particlesystem.h"

//// Function Pointers for extentions
int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance, 
				   LPSTR lpCmdLine, int nCmdShow);
LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

//// Declarations
void init(void);
void render(void);
void shutDown(void);
void initParticleSystem();

//// Globals
HWND   g_hWnd      = NULL;
HDC	   g_hDC       = NULL;
HGLRC  g_hRC       = NULL;

float g_fSpinX = 0.0f;
float g_fSpinY = 0.0f;

// Our variables used to keep track of delta time
// Windows2k and WindowsXP have a resolution of 10ms, so if this would be an issue we could look at QueryPerformanceCounter
double curTime = 0;
double lastTime= 0;

// A placeholder for our particlesystems. I've set a limit for 10 systems here in this example.
ParticleSystem *pSys[10];

// What system we're going to show.
int system = 0;
// How many particlesystems are in use at the same time. Our renderer uses this for looping through the active ones.
int systemsInUse = 0;

// The programs entrypoint
// As it is now, this example is a Win32 Application, though the ParticleSystem code should be cross plattform.
int WINAPI WinMain(	HINSTANCE hInstance,
					HINSTANCE hPrevInstance,
					LPSTR     lpCmdLine,
					int       nCmdShow )
{
	WNDCLASSEX winClass; 
	MSG        uMsg;

    memset(&uMsg,0,sizeof(uMsg));

	winClass.lpszClassName = "GRAF_DAT_PARTICLES";
	winClass.cbSize        = sizeof(WNDCLASSEX);
	winClass.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	winClass.lpfnWndProc   = WindowProc;
	winClass.hInstance     = hInstance;
    winClass.hIcon	       = LoadIcon(hInstance, (LPCTSTR)IDB_BITMAP1);
    winClass.hIconSm	   = LoadIcon(hInstance, (LPCTSTR)IDB_BITMAP1);
	winClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
	winClass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	winClass.lpszMenuName  = NULL;
	winClass.cbClsExtra    = 0;
	winClass.cbWndExtra    = 0;
	
	if( !RegisterClassEx(&winClass) )
		return E_FAIL;

	g_hWnd = CreateWindowEx( NULL, "GRAF_DAT_PARTICLES", 
							"Graf.Dat - PartikkelSystem",
						     WS_OVERLAPPEDWINDOW | WS_VISIBLE,
					         0, 0, 1024, 786, NULL, NULL, hInstance, NULL );

	if( g_hWnd == NULL )
		return E_FAIL;

    ShowWindow( g_hWnd, nCmdShow );
    UpdateWindow( g_hWnd );

	init();

	while( uMsg.message != WM_QUIT )
	{
		if( PeekMessage( &uMsg, NULL, 0, 0, PM_REMOVE ) )
		{
			TranslateMessage( &uMsg );
			DispatchMessage( &uMsg );
		}
        else
			
			curTime = timeGetTime();
			render();
			lastTime = curTime;
	}

	if( g_hRC != NULL )
	{
		wglMakeCurrent( NULL, NULL );
		wglDeleteContext( g_hRC );
		g_hRC = NULL;
	}

	if( g_hDC != NULL )
	{
		ReleaseDC( g_hWnd, g_hDC );
		g_hDC = NULL;
	}

    UnregisterClass( "GRAF_DAT_PARTICLES", hInstance );

	return uMsg.wParam;
}

LRESULT CALLBACK WindowProc( HWND   hWnd, 
							 UINT   msg, 
							 WPARAM wParam, 
							 LPARAM lParam )
{
    static POINT ptLastMousePosit;
	static POINT ptCurrentMousePosit;
	static bool bMousing;

    switch( msg )
	{
        case WM_KEYDOWN:
		{
			switch( wParam )
			{
				case VK_ESCAPE:
					PostQuitMessage(0);
				break;

				case VK_SPACE:
					system++;
					initParticleSystem();
				break;

			}
		}
        break;

        case WM_LBUTTONDOWN:
		{
			ptLastMousePosit.x = ptCurrentMousePosit.x = LOWORD (lParam);
            ptLastMousePosit.y = ptCurrentMousePosit.y = HIWORD (lParam);
			bMousing = true;
		}
		break;

		case WM_LBUTTONUP:
		{
			bMousing = false;
		}
		break;

		case WM_MOUSEMOVE:
		{
			ptCurrentMousePosit.x = LOWORD (lParam);
			ptCurrentMousePosit.y = HIWORD (lParam);

			if( bMousing )
			{
				g_fSpinX -= (ptCurrentMousePosit.x - ptLastMousePosit.x);
				g_fSpinY -= (ptCurrentMousePosit.y - ptLastMousePosit.y);
			}
			
			ptLastMousePosit.x = ptCurrentMousePosit.x;
            ptLastMousePosit.y = ptCurrentMousePosit.y;
		}
		break;

        case WM_SIZE:
		{
			int nWidth  = LOWORD(lParam); 
			int nHeight = HIWORD(lParam);
			glViewport(0, 0, nWidth, nHeight);

			glMatrixMode( GL_PROJECTION );
			glLoadIdentity();
			gluPerspective( 45.0, (GLdouble)nWidth / (GLdouble)nHeight, 0.1, 1000.0);
		}
		break;

		case WM_CLOSE:
		{
			PostQuitMessage(0);	
		}

        case WM_DESTROY:
		{
            PostQuitMessage(0);
		}
        break;
		
		default:
		{
			return DefWindowProc( hWnd, msg, wParam, lParam );
		}
		break;
	}

	return 0;
}

void init( void )
{
	GLuint PixelFormat;

	// Set up the windows part
	PIXELFORMATDESCRIPTOR pfd;
	memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));

    pfd.nSize      = sizeof(PIXELFORMATDESCRIPTOR);
    pfd.nVersion   = 1;
    pfd.dwFlags    = PFD_DRAW_TO_WINDOW |PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 16;
    pfd.cDepthBits = 16;
	
	g_hDC = GetDC( g_hWnd );
	PixelFormat = ChoosePixelFormat( g_hDC, &pfd );
	SetPixelFormat( g_hDC, PixelFormat, &pfd);
	g_hRC = wglCreateContext( g_hDC );
	wglMakeCurrent( g_hDC, g_hRC );

	// Init 
	glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
	glEnable(GL_TEXTURE_2D);

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, 800.0f / 600.0f, 0.1f, 100.0f);

	initParticleSystem();

}

void initParticleSystem()
{
	for (int i = 0; pSys[i] != NULL && i < sizeof(pSys); i++)
	{
		delete pSys[i];
		pSys[i] = NULL;
	}
	
	int curSys = 0;
	switch (system)
	{
	case 0:
		systemsInUse = 1;
		pSys[curSys] = new ParticleSystem();

		pSys[curSys]->setTexture( ".\\snow.bmp" );
		pSys[curSys]->setParticleCount (3000 );
		pSys[curSys]->setReleaseNumber( 20 );
		pSys[curSys]->setReleaseInterval( 0.05f );
		pSys[curSys]->setParticleLife( 6.5f );
		pSys[curSys]->setFade(0.9f);
		pSys[curSys]->setParticleSize( 30.0f );
		pSys[curSys]->setInitialPosition( vector3f( -5.0f, 4.0f, -5.0f) );
		pSys[curSys]->setBoxCorner( vector3f( 5.0f, 6.0f, 5.0f));
		pSys[curSys]->setInitialVelocity( vector3f( 0.0f, -3.0f, 0.0f) );
		pSys[curSys]->setGravity( vector3f( 0.0f, -1.0f, 0.0f) );
		pSys[curSys]->setWind( vector3f( 0.0f, 1.8f, 0.0f) );
		pSys[curSys]->setWindState(true);
	break;

	case 1:
		systemsInUse = 2;
		pSys[curSys] = new ParticleSystem();

		pSys[curSys]->setTexture( ".\\fire.bmp" );
		pSys[curSys]->setParticleCount (2000 );
		pSys[curSys]->setReleaseNumber( 20 );
		pSys[curSys]->setReleaseInterval( 0.05f );
		pSys[curSys]->setParticleLife( 2.0f );
		pSys[curSys]->setFade(0.8f);
		pSys[curSys]->setParticleSize( 50.0f );
		pSys[curSys]->setInitialPosition( vector3f( -1.5f, -2.0f, -0.5f) );
		pSys[curSys]->setBoxCorner( vector3f( 1.5f, -2.0f, 1.0f));

		pSys[curSys]->setInitialVelocity( vector3f( 0.0f, 1.0f, 0.0f) );
		pSys[curSys]->setGravity( vector3f( 0.0f, 1.0f, 0.0f) );
		pSys[curSys]->setWind( vector3f( 0.0f, 0.0f, 0.0f) );
		pSys[curSys]->init();

		curSys++;
		pSys[curSys] = new ParticleSystem();
		pSys[curSys]->setTexture( ".\\smoke.bmp" );
		pSys[curSys]->setParticleCount (2000 );
		pSys[curSys]->setReleaseNumber( 14 );
		pSys[curSys]->setReleaseInterval( 0.1f );
		pSys[curSys]->setParticleLife( 4.5f );
		pSys[curSys]->setFade(0.8f);
		pSys[curSys]->setParticleSize( 70.0f );
		pSys[curSys]->setInitialPosition( vector3f( -1.5f, -1.8f, -0.5f) );
		pSys[curSys]->setBoxCorner( vector3f( 1.5f, -1.8f, 1.0f));

		pSys[curSys]->setInitialVelocity( vector3f( 0.0f, 0.7f, 0.0f) );
		pSys[curSys]->setGravity( vector3f( 0.0f, 0.8f, 0.0f) );
		pSys[curSys]->setWind( vector3f( 0.0f, 0.0f, 0.0f) );
		pSys[curSys]->init();
	break;

	case 2:
		systemsInUse = 1;
		pSys[0] = new ParticleSystem();

		pSys[0]->setTexture( ".\\smoke.bmp" );
		pSys[0]->setParticleCount (1500 );
		pSys[0]->setReleaseNumber( 5 );
		pSys[0]->setReleaseInterval( 0.01f );
		pSys[0]->setParticleLife( 2.5f );
		pSys[0]->setFade(0.8f);
		pSys[0]->setParticleSize( 30.0f );
		pSys[0]->setInitialPosition( vector3f( 0.0f, -2.0f, 0.0f) );
		pSys[0]->setBoxCorner( vector3f( 1.0f, -2.0f, 1.0f));

		pSys[0]->setInitialVelocity( vector3f( 0.0f, 1.0f, 0.0f) );
		pSys[0]->setGravity( vector3f( 0.0f, 1.0f, 0.0f) );
		pSys[0]->setWind( vector3f( 0.6f, 0.0f, 0.0f) );
		pSys[0]->setWindState(true);
	break;
	case 3:
		systemsInUse = 1;
		pSys[curSys] = new ParticleSystem();

		pSys[curSys]->setTexture( ".\\fire.bmp" );
		pSys[curSys]->setParticleCount (1500 );
		pSys[curSys]->setReleaseNumber( 5 );
		pSys[curSys]->setReleaseInterval( 0.01f );
		pSys[curSys]->setParticleLife( 2.5f );
		pSys[curSys]->setFade(0.8f);
		pSys[curSys]->setParticleSize( 30.0f );
		pSys[curSys]->setRandomFactor(1.0f);
		pSys[curSys]->setRandomOn(true);

		pSys[curSys]->setInitialPosition( vector3f( 0.0f, -2.0f, 2.0f) );

		pSys[curSys]->setInitialVelocity( vector3f( 0.0f, 1.0f, 0.0f) );
		pSys[curSys]->setGravity( vector3f( 0.0f, 1.0f, 0.0f) );
		pSys[curSys]->setWind( vector3f( 0.6f, 0.0f, 0.0f) );
		pSys[curSys]->setWindState(true);
	break;
	default:
		system = 0;
		initParticleSystem();
		break;
	}
}

void render( void )
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
	glTranslatef( 0.0f, 0.0f, -10.0f );
    glRotatef( -g_fSpinY, 1.0f, 0.0f, 0.0f );
    glRotatef( -g_fSpinX, 0.0f, 1.0f, 0.0f );

	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE );

	double delta = curTime - lastTime;
	for (int i = 0; i < systemsInUse; i++)
	{
		pSys[i]->update(delta);
		pSys[i]->render();
	}

	SwapBuffers( g_hDC );
}
